from typing import Dict, Optional

from compressed_tensors import CompressionFormat, SparsityCompressionConfig
from compressed_tensors.quantization.utils import is_model_quantized
from torch import Tensor
from torch.nn import Module

from llmcompressor.core import active_session
from llmcompressor.pytorch.utils import ModuleSparsificationInfo
from llmcompressor.transformers.compression.helpers import (
    infer_sparsity_structure_from_model,
    infer_sparsity_structure_from_stage_modifiers,
)


class SparsityConfigMetadata:
    """
    Class of helper functions for filling out a SparsityCompressionConfig with readable
    metadata from the model
    """

    @staticmethod
    def infer_global_sparsity(
        model: Module, state_dict: Optional[Dict[str, Tensor]] = None
    ) -> float:
        """
        Calculates the global percentage of sparse zero weights in the model

        :param model: pytorch model to infer sparsity of
        :param state_dict: optional state_dict to replace that in model, used for
        gathering global FSDP model info
        :return: global sparsity of model
        """

        info = ModuleSparsificationInfo(model, state_dict=state_dict)
        global_sparsity = info.params_sparse_percent / 100.0  # convert % to float
        return global_sparsity

    @staticmethod
    def infer_sparsity_structure(model: Optional[Module] = None) -> str:
        """
        Determines what sparsity structure, if any, was applied.

        First, there is an attempt to dedue the sparsity structure
        from the currently active sparse session.

        If that fails, the sparsity structure is inferred from the
        model (if provided)

        Finally, if both fail, the sparsity structure is set to
        "unstructured"

        :return: sparsity structure as a string
        """
        sparsity_structure = None

        current_session = active_session()
        stage_modifiers = current_session.lifecycle.modifiers
        if stage_modifiers:
            sparsity_structure = infer_sparsity_structure_from_stage_modifiers(
                stage_modifiers
            )

        if model and sparsity_structure is None:
            sparsity_structure = infer_sparsity_structure_from_model(model)

        return sparsity_structure or "unstructured"

    @staticmethod
    def from_pretrained(
        model: Module,
        state_dict: Optional[Dict[str, Tensor]] = None,
        compress: bool = False,
    ) -> Optional["SparsityCompressionConfig"]:
        """
        Determines compression type and informational parameters for a given model

        :param model: pytorch model to calculate sparsity config for
        :param state_dict: optional state_dict to replace that in model, used for
        gathering global FSDP model info
        :param compress: whether or not to compress the model on disk
        :return: compression config inferred from the model
        """

        global_sparsity = SparsityConfigMetadata.infer_global_sparsity(
            model, state_dict=state_dict
        )

        if global_sparsity < 0.05:
            return None

        sparsity_structure = SparsityConfigMetadata.infer_sparsity_structure(
            model=model
        )
        if is_model_quantized(model):
            # compressing a sparse quantized model is not supported yet
            format = CompressionFormat.dense.value
        elif compress:
            format = CompressionFormat.sparse_bitmask.value
        else:
            format = CompressionFormat.dense.value

        return SparsityCompressionConfig.load_from_registry(
            format,
            global_sparsity=global_sparsity,
            sparsity_structure=sparsity_structure,
        )

    @staticmethod
    def fill_config_details(
        config: SparsityCompressionConfig,
        model: Module,
        state_dict: Optional[Dict[str, Tensor]] = None,
    ):
        """
        Fills in informational sparsity parameters from a given model

        :param config: sparsity config to fill in
        :param model: pytorch model to infer config parameters from
        :param state_dict: optional state_dict to replace that in model, used for
        gathering global FSDP model info
        """
        config.global_sparsity = SparsityConfigMetadata.infer_global_sparsity(
            model, state_dict=state_dict
        )
        config.sparsity_structure = SparsityConfigMetadata.infer_sparsity_structure()



"""
Module for defining and managing callback event lifecycles in the LLM Compressor.

This module provides a class for defining event lifecycles when callbacks
are used to communicate the state of the training pipeline.
"""

from typing import List

from loguru import logger

from llmcompressor.core.events.event import Event, EventType
from llmcompressor.core.events.event_lifecycle import EventLifecycle

__all__ = [
    "CallbacksEventLifecycle",
]


class CallbacksEventLifecycle(EventLifecycle):
    """
    An event lifecycle for when callbacks are used to communicate the
    state of the training pipeline.

    The expected lifecycle is as follows with optional gradient accumulation:
        for gradient_batches in training loop:
            for batch in gradient_batches or [gradient_batches]:
                batch_start() -> [BATCH_START]
                loss_calculated() -> [LOSS_CALCULATED]

                if not last batch:
                    batch_end -> [BATCH_END]
                else:
                    optim_pre_step() -> [OPTIM_PRE_STEP]
                    optim_post_step() -> [OPTIM_POST_STEP]
                    batch_end -> [BATCH_END]

    Which gives the following logic:
        - BATCH_START: must be called first or after OPTIM_POST_STEP
        - LOSS_CALCULATED: must be called after BATCH_START and
            before BATCH_END or OPTIM_POST_STEP
        - OPTIM_PRE_STEP: must be called after LOSS_CALCULATED
            and before OPTIM_POST_STEP
        - OPTIM_POST_STEP: must be called after OPTIM_PRE_STEP and before BATCH_END
        - BATCH_END: must be called after LOSS_CALCULATED or OPTIM_POST_STEP
    """

    def __init__(self, type_first: EventType, start: Event):
        """
        Initialize the CallbacksEventLifecycle.

        :param type_first: The first event type to be called
        :type type_first: EventType
        :param start: The start event to base the lifecycle off of
        :type start: Event
        """
        super().__init__(type_first=type_first, start=start)
        self.skip_post_step = False

    def batch_start_events(self) -> List[Event]:
        """
        Return the list of events to be called for the batch start.

        :return: The list of events to be called for the batch start
        :rtype: List[Event]
        :raises ValueError: If batch start is not called first or if it
            is not called after batch end
        """
        if self.type_first != EventType.BATCH_START:
            logger.error("batch start must be called first for callbacks")
            raise ValueError("batch start must be called first for callbacks")

        if self.last_type not in {None, EventType.BATCH_END, EventType.OPTIM_POST_STEP}:
            logger.error("batch start must be called after batch end")
            raise ValueError("batch start must be called after batch end")

        self.last_type = EventType.BATCH_START
        step_ready = self.check_batches_per_step_count(increment=True)
        logger.debug(
            "Batch start event processed with step_ready={}, "
            "global_step={}, and global_batch={}",
            step_ready,
            self.global_step,
            self.global_batch,
        )

        return [self.new_instance(type_=EventType.BATCH_START)]

    def loss_calculated_events(self) -> List[Event]:
        """
        Return the list of events to be called for the loss calculated.

        :return: The list of events to be called for the loss calculated
        :rtype: List[Event]
        :raises ValueError: If loss calculated is not called after batch start
        """
        if self.last_type != EventType.BATCH_START:
            logger.error("loss calculated must be called after batch start")
            raise ValueError("loss calculated must be called after batch start")

        self.last_type = EventType.LOSS_CALCULATED
        logger.debug(
            "Loss calculated event processed with global_batch={} and global_step={}",
            self.global_batch,
            self.global_step,
        )

        return [self.new_instance(type_=EventType.LOSS_CALCULATED)]

    def optim_pre_step_events(self) -> List[Event]:
        """
        Return the list of events to be called for the optim pre step.

        :return: The list of events to be called for the optim pre step
        :rtype: List[Event]
        :raises ValueError: If optim pre step is not called after batch start
            or loss calculated
        """
        if self.last_type not in {EventType.LOSS_CALCULATED}:
            logger.error("optim pre step must be called after loss calculated")
            raise ValueError("optim pre step must be called after loss calculated")

        self.last_type = EventType.OPTIM_PRE_STEP
        at_invocation = self.check_invocations_per_step_count(increment=True)
        logger.debug(
            "Optim pre step event processed with at_invocation={}, "
            "global_step={}, and global_batch={}",
            at_invocation,
            self.global_step,
            self.global_batch,
        )

        if not at_invocation:
            self.skip_post_step = True
            return []
        else:
            self.skip_post_step = False
            return [self.new_instance(type_=EventType.OPTIM_PRE_STEP)]

    def optim_post_step_events(self) -> List[Event]:
        """
        Return the list of events to be called for the optim post step.

        :return: The list of events to be called for the optim post step
        :rtype: List[Event]
        :raises ValueError: If optim post step is not called after optim pre step
        """
        if self.last_type != EventType.OPTIM_PRE_STEP:
            logger.error("optim post step must be called after optim pre step")
            raise ValueError("optim post step must be called after optim pre step")

        self.last_type = EventType.OPTIM_POST_STEP
        logger.debug(
            "Optim post step event processed with global_batch={} and global_step={}",
            self.global_batch,
            self.global_step,
        )

        if self.skip_post_step:
            return []
        else:
            return [self.new_instance(type_=EventType.OPTIM_POST_STEP)]

    def batch_end_events(self) -> List[Event]:
        """
        Return the list of events to be called for the batch end.

        :return: The list of events to be called for the batch end
        :rtype: List[Event]
        :raises ValueError: If batch end is not called after optim post step,
            loss calculated, or batch start
        """
        if self.last_type not in {
            EventType.OPTIM_POST_STEP,
            EventType.LOSS_CALCULATED,
        }:
            logger.error(
                "batch end must be called after loss calculated or optim post step"
            )
            raise ValueError(
                "batch end must be called after loss calculated or optim post step"
            )

        self.last_type = EventType.BATCH_END
        logger.debug(
            "Batch end event processed with global_batch={} and global_step={}",
            self.global_batch,
            self.global_step,
        )

        return [self.new_instance(type_=EventType.BATCH_END)]



from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Union

from llmcompressor.modifiers import Modifier
from llmcompressor.recipe.recipe import Recipe, RecipeTuple

__all__ = ["RecipeContainer"]


@dataclass
class RecipeContainer:
    """
    A container for recipes to be used in a session. Provides utilities
    to update the recipes and compile them into a single recipe.

    :param compiled_recipe: the compiled recipe from the recipes list
    :param recipes: the list of RecipeTuple instances to be compiled
    :param applied_stages: list of recipe stages that have already been applied
    """

    compiled_recipe: Optional[Recipe] = None
    recipes: List[RecipeTuple] = field(default_factory=list)
    applied_stages: List[str] = field(default_factory=list)

    def update(
        self,
        recipe: Union[
            str, List[str], Recipe, List[Recipe], Modifier, List[Modifier], None
        ] = None,
        recipe_stage: Union[str, List[str], List[List[str]], None] = None,
        recipe_args: Union[Dict[str, Any], List[Dict[str, Any]], None] = None,
        **kwargs,
    ) -> Dict:
        """
        Update the recipes in the container. If a recipe is provided, it will
        reset any existing compiled_recipe in the container. Must call
        `check_compile_recipe` to re-compile the recipes into a single compiled_recipe.
        If no recipe is provided, does nothing and returns the kwargs.

        Can provide multiple recipes to update the container with:
        >>> container = RecipeContainer()
        >>> recipe_str_1 = '''
        ... test_stage:
        ...     pruning_modifiers:
        ...         ConstantPruningModifier:
        ...             start: 0.0
        ...             end: 2.0
        ...             targets: ['re:.*weight']
        ... '''
        >>> recipe_str_2 = '''
        ... test_stage:
        ...     pruning_modifiers:
        ...         ConstantPruningModifier:
        ...             start: 3.0
        ...             end: 4.0
        ...             targets: ['re:.*weight']
        ... '''
        >>> result = container.update(recipe=[recipe_str_1, recipe_str_2])

        :param recipe: the recipe to update the container with
        :param recipe_stage: the recipe stage to update the container with
        :param recipe_args: the recipe args to update the recipe with
        :param kwargs: additional kwargs to return
        :return: the passed in kwargs
        """
        if recipe is None or isinstance(recipe, list) and len(recipe) == 0:
            return kwargs

        self.compiled_recipe = None

        if isinstance(recipe, Modifier) or (
            isinstance(recipe, list)
            and all(isinstance(mod, Modifier) for mod in recipe)
        ):
            recipe = Recipe.create_instance(recipe)

        if not isinstance(recipe, list):
            recipe = [recipe]

        if recipe_stage is None:
            recipe_stage = [None] * len(recipe)
        else:
            if not isinstance(recipe_stage, list):
                recipe_stage = [[recipe_stage]] * len(recipe)
            if not isinstance(recipe_stage[0], list):
                recipe_stage = [recipe_stage] * len(recipe)

        if recipe_args is None:
            recipe_args = [{}] * len(recipe)
        elif not isinstance(recipe_args, list):
            recipe_args = [recipe_args] * len(recipe)

        if len(recipe) != len(recipe_stage) or len(recipe) != len(recipe_args):
            raise ValueError(
                "recipe, recipe_stage, and recipe_args must be the same length"
            )

        for rec, stage, args in zip(recipe, recipe_stage, recipe_args):
            if isinstance(rec, str):
                rec = Recipe.create_instance(rec)
            self.recipes.append(RecipeTuple(rec, stage, args))

        return kwargs

    def update_applied_stages(self, new_stages: List[str]):
        """
        Updates the applied_stages list with new stages, indicating their structure
        has already been applied

        :param new_stages: new stage names to add
        """
        for stage in new_stages:
            if stage not in self.applied_stages:
                self.applied_stages.append(stage)

    def check_compile_recipe(self) -> bool:
        """
        Check if the recipes need to be compiled into a single recipe and
        compile them if they do.

        :return: True if the recipes were compiled, False otherwise
        """
        if self.compiled_recipe is None and self.recipes:
            self.compiled_recipe = Recipe.simplify_combine_recipes(self.recipes)
            return True

        return False

    def check_any_recipe_exists(self) -> bool:
        """
        Checks if any recipes have been added to the container, compiled or not

        :return: True if any recipes exist in the container, False otherwise
        """
        if self.compiled_recipe is not None:
            return True
        if len(self.recipes) > 0:
            return True

        return False



import inspect
import logging
from pathlib import Path
from typing import Optional, Union

import torch
from accelerate import load_checkpoint_and_dispatch
from compressed_tensors.compressors import ModelCompressor
from compressed_tensors.quantization import (
    QuantizationStatus,
    apply_quantization_config,
)
from loguru import logger
from torch.nn import Module
from transformers import AutoModelForCausalLM, PreTrainedModel

from llmcompressor.pytorch.model_load.helpers import initialize_recipe
from llmcompressor.transformers.sparsification.compressed_tensors_utils import (
    modify_save_pretrained,
)
from llmcompressor.transformers.utils.helpers import (
    download_model_directory,
    resolve_recipe,
)

__all__ = ["SparseAutoModel", "SparseAutoModelForCausalLM", "get_shared_tokenizer_src"]


class SparseAutoModelForCausalLM(AutoModelForCausalLM):
    """
    LLM Compressor wrapper for the AutoModelForCausalLM class
    Its lifecycle is defined as follows:
    1. If pretrained_model_name_or_path is a HuggingFace stub
       the appropriate HuggingFace model will be downloaded
       (if required) and the path to the deployment directory
       of the model will be retrieved
    2. The original model definition will be loaded, without
        the model weights
    3. The appropriate recipe will be applied to the model
       if requested or required
    4. The appropriate set of weights will be loaded into the model
    """

    @classmethod
    def from_pretrained(
        cls,
        pretrained_model_name_or_path,
        run_compressed: bool = False,
        recipe: Optional[Union[str, Path]] = None,
        *model_args,
        **kwargs,
    ) -> PreTrainedModel:
        """
         A wrapper around the AutoModelForCausalLM.from_pretrained method

        :param pretrained_model_name_or_path: the name of or path to the model to load
        :param recipe: the path to the recipe file to apply to the model. Can be a
            string or Path object. If None, a recipe will be searched for in the
            pretrained_model_name_or_path directory and applied if found
        :return the created model for causal language modeling
        """

        def skip(*args, **kwargs):
            pass

        # Skip the initializer step. This accelerates the loading
        # of the models, especially for the quantized models
        torch.nn.init.kaiming_uniform_ = skip
        torch.nn.init.uniform_ = skip
        torch.nn.init.normal_ = skip

        pretrained_model_name_or_path = (
            pretrained_model_name_or_path.as_posix()
            if isinstance(pretrained_model_name_or_path, Path)
            else pretrained_model_name_or_path
        )

        pretrained_model_name_or_path = download_model_directory(
            pretrained_model_name_or_path, **kwargs
        )

        # instantiate compressor from model config
        compressor = ModelCompressor.from_pretrained(
            pretrained_model_name_or_path, **kwargs
        )

        # temporarily set the log level to error, to ignore printing out long missing
        # and unexpected key error messages (these are EXPECTED for quantized models)
        transformers_logger = logging.getLogger("transformers.modeling_utils")
        restore_log_level = transformers_logger.getEffectiveLevel()
        transformers_logger.setLevel(level=logging.ERROR)

        if kwargs.get("trust_remote_code"):
            # By artifically aliasing
            # class name SparseAutoModelForCausallLM to
            # AutoModelForCausalLM we can "trick" the
            # `from_pretrained` method into properly
            # resolving the logic when
            # (has_remote_code and trust_remote_code) == True
            cls.__name__ = AutoModelForCausalLM.__name__

        model = super(AutoModelForCausalLM, cls).from_pretrained(
            pretrained_model_name_or_path, *model_args, **kwargs
        )

        if model.dtype != model.config.torch_dtype:
            logger.warning(
                f"The dtype of the loaded model: {model.dtype} is different "
                "from from the dtype specified in the model config: "
                f"{model.config.torch_dtype}."
                "To load the model in the format that it was previously saved in, "
                "set torch_dtype=`auto` in the SparseAutoModel creation call."
            )

        # restore transformers logging level now that model shell is loaded
        transformers_logger.setLevel(level=restore_log_level)

        # HfQuantizer Quantization
        if hasattr(model.config, "quantization_config"):
            return model

        # override the PreTrainedModel instance with compression save function
        modify_save_pretrained(model)

        # If model is quantized or compressed on disk, initialize quantization
        # structure and run decompression
        if compressor is not None:
            quantization_config = compressor.quantization_config
            is_compressed = (
                quantization_config is not None
                and quantization_config.quantization_status
                == QuantizationStatus.COMPRESSED
            )
            if run_compressed and is_compressed:
                # initialize quantization, don't decompress
                apply_quantization_config(
                    model, quantization_config, run_compressed=True
                )
                model = load_checkpoint_and_dispatch(
                    model, pretrained_model_name_or_path
                )
            else:
                # initialize quantization and decompress weights
                if quantization_config is not None:
                    quantization_config.quantization_status = QuantizationStatus.FROZEN
                compressor.decompress(
                    model_path=pretrained_model_name_or_path, model=model
                )
        recipe = resolve_recipe(recipe=recipe, model_path=pretrained_model_name_or_path)

        if recipe:
            initialize_recipe(model=model, recipe_path=recipe)

        return model


class SparseAutoModel:
    """
    Factory class for creating sparse models using transformers AutoModel classes
    """

    @staticmethod
    def text_generation_from_pretrained(
        model_name_or_path: str,
        sequence_length: Optional[int] = None,
        recipe: Optional[Union[str, Path]] = None,
        trust_remote_code: bool = False,
        torch_dtype: Union[str, torch.dtype] = "auto",
        **kwargs,
    ) -> Module:
        """
        :param model_name_or_path: the name of or path to the model to load
        :param sequence_length: the maximum length of the sequence to generate.
            If None, will use the default sequence length for the model.
            Defaults to None.
        :param recipe: the recipe to apply to the model. If None, no recipe is applied
        :param trust_remote_code: related to trust_remote_code in HF transformers.
            If True, will execute the modelling code from the model directory
            (if present). Defaults to False.
        :param torch_dtype: the torch dtype to use for the model. If "auto", will
            use the default dtype for the model. Defaults to "auto".
        :return: the created model for text generation
        """

        model = SparseAutoModelForCausalLM.from_pretrained(
            model_name_or_path,
            torch_dtype=torch_dtype,
            trust_remote_code=trust_remote_code,
            recipe=recipe,
            **kwargs,
        )
        if sequence_length is not None:
            model.seqlen = sequence_length

        return model


def get_shared_tokenizer_src(student: Module, teacher: Optional[Module]) -> str:
    """
    Get a tokenizer source used for both student and teacher, assuming
    that they could be shared

    :param student: the student model
    :param teacher: the teacher model
    :return: the source for the tokenizer shared between teacher and model
    """

    if teacher is not None and teacher not in ("disable", "self"):
        student_forward_params = list(
            inspect.signature(student.forward).parameters.keys()
        )
        teacher_forward_params = list(
            inspect.signature(teacher.forward).parameters.keys()
        )
        diff = [p for p in student_forward_params if p not in teacher_forward_params]
        if diff:
            raise RuntimeError(
                "Teacher tokenizer cannot be used for student "
                f"due to missing args: {diff}"
            )
        src_model = teacher
    else:
        src_model = student
    return src_model.config._name_or_path



from typing import Optional, Union

from compressed_tensors.registry import RegistryMixin
from datasets import Dataset, IterableDataset
from loguru import logger
from transformers import AutoTokenizer

from llmcompressor.transformers.finetune.data.data_args import DataTrainingArguments
from llmcompressor.transformers.finetune.data.data_helpers import (
    LABELS_MASK_VALUE,
    get_custom_datasets_from_path,
    get_raw_dataset,
)


class TextGenerationDataset(RegistryMixin):
    """
    Base class for text datasets, handles tokenization and dataset splits

    :param text_column: name of column corresponding to text in the dataset
    :param data_args: configuration settings for dataset loading
    :param split: split from dataset to load, for instance `test` or `train[:5%]`
    :param tokenizer: tokenizer to use on dataset
    """

    PROMPT_KEY = "prompt"

    def __init__(
        self,
        text_column: str,
        data_args: DataTrainingArguments,
        split: str,
        tokenizer: AutoTokenizer,
    ):
        self.text_column = text_column
        self.tokenizer = tokenizer
        self.data_args = data_args
        self.raw_kwargs = data_args.raw_kwargs or {}
        self.split = split
        self.dvc_dataset = (
            True if self.data_args.dvc_data_repository is not None else False
        )
        self.custom_dataset = True if self.data_args.dataset_path is not None else False

        # configure padding
        if data_args.concatenate_data:
            self.padding = False
        elif data_args.pad_to_max_length:
            self.padding = "max_length"
        else:
            self.padding = False

        if self.tokenizer:
            if not self.tokenizer.pad_token:
                self.tokenizer.pad_token = self.tokenizer.eos_token

        # configure sequence length
        max_seq_length = data_args.max_seq_length
        model_max_length = tokenizer.model_max_length if tokenizer else max_seq_length
        if self.tokenizer and max_seq_length > model_max_length:
            logger.warning(
                f"The max_seq_length passed ({max_seq_length}) is larger than "
                f"the maximum length for the model ({tokenizer.model_max_length}). "
                f"Using max_seq_length={tokenizer.model_max_length}."
            )
        self.max_seq_length = min(data_args.max_seq_length, model_max_length)

    def get_raw_dataset(self, cache_dir: Optional[str] = None) -> Dataset:
        """
        Load the raw dataset from Hugging Face, using cached copy if available

        :param cache_dir: disk location to search for cached dataset
        :return: the requested dataset
        """
        if self.custom_dataset:
            if self.dvc_dataset:
                self.raw_kwargs["storage_options"] = {
                    "url": self.data_args.dvc_data_repository
                }
                self.raw_kwargs["data_files"] = self.data_args.dataset_path
            else:
                self.raw_kwargs["data_files"] = get_custom_datasets_from_path(
                    self.data_args.dataset_path,
                    self.data_args.dataset
                    if hasattr(self.data_args, "dataset")
                    else self.data_args.dataset_name,
                )

        return get_raw_dataset(
            self.data_args,
            cache_dir,
            split=self.split,
            streaming=self.data_args.streaming,
            **self.raw_kwargs,
        )

    def tokenize_and_process(
        self, raw_dataset: Optional[Dataset] = None, add_labels: Optional[bool] = True
    ) -> Dataset:
        """
        Sets up the raw dataset for finetuning, performs tokenization, concatenates
        entries to max sequence length if desired, and adds labels to each entry

        :param raw_dataset: dataset to process
        :param add_labels: whether to include labels in tokenized output
        """

        # helper fn for tokenizing text column
        def tokenize_fn(data):
            result = self.tokenizer(
                data[self.text_column],
                padding=self.padding,
                max_length=self.max_seq_length,
                truncation=True,
            )

            # store unpadded prompt so we can mask out correct number of elements
            # in the labels
            if self.PROMPT_KEY in data:
                result[self.PROMPT_KEY] = self.tokenizer(
                    data[self.PROMPT_KEY],
                    max_length=self.max_seq_length,
                    truncation=True,
                )["input_ids"]

            return result

        # helper fn for filling to max_sequence_length by concatenating entries
        def group_text_fn(data):
            concatenated_data = {k: sum(data[k], []) for k in data.keys()}
            total_length = len(concatenated_data[list(data.keys())[0]])
            total_length = (total_length // self.max_seq_length) * self.max_seq_length
            result = {
                k: [
                    t[i : i + self.max_seq_length]
                    for i in range(0, total_length, self.max_seq_length)
                ]
                for k, t in concatenated_data.items()
            }
            return result

        # helper fn for adding labels, needed for loss calculation
        def label_fn(data):
            # if the dataset uses prompts, mask them out so they don't contribute
            # to the loss calculation
            prompt_len = 0
            if self.PROMPT_KEY in data:
                prompt_len = len(data[self.PROMPT_KEY])
            data["labels"] = data["input_ids"].copy()
            data["labels"][:prompt_len] = [LABELS_MASK_VALUE] * prompt_len

            # mask out padding in the labels as well
            padding = len(data["attention_mask"]) - sum(data["attention_mask"])
            if padding > 0:
                data["labels"][-padding:] = [LABELS_MASK_VALUE] * padding
            return data

        if raw_dataset is None:
            raw_dataset = self.get_raw_dataset()

        dataset = self.map(
            raw_dataset,
            function=tokenize_fn,
            batched=True,
            remove_columns=[self.text_column],
            num_proc=self.data_args.preprocessing_num_workers,
            load_from_cache_file=not self.data_args.overwrite_cache,
            desc="Running tokenizer on dataset",
        )

        if self.data_args.concatenate_data:
            dataset = self.map(
                dataset,
                function=group_text_fn,
                batched=True,
                num_proc=self.data_args.preprocessing_num_workers,
                load_from_cache_file=not self.data_args.overwrite_cache,
                desc="Grouping text",
            )

        if isinstance(dataset, IterableDataset):
            # so we can get column names from streamed_datasets
            dataset = dataset._resolve_features()

        column_names = dataset.column_names
        if isinstance(column_names, dict):
            column_names = column_names[list(column_names)[0]]

        if add_labels:
            dataset = self.map(
                dataset,
                function=label_fn,
                batched=False,  # not compatible with batching, need row lengths
                remove_columns=[self.PROMPT_KEY]
                if self.PROMPT_KEY in column_names
                else None,
                num_proc=self.data_args.preprocessing_num_workers,
                load_from_cache_file=not self.data_args.overwrite_cache,
                desc="Adding labels",
            )
        else:
            dataset = self.map(
                dataset,
                batched=False,  # not compatible with batching, need row lengths
                remove_columns=[self.PROMPT_KEY]
                if self.PROMPT_KEY in column_names
                else None,
            )

        return dataset

    def map(
        self, dataset: Union[Dataset, IterableDataset], **kwargs
    ) -> Union[Dataset, IterableDataset]:
        """
        Wrapper function around Dataset.map and IterableDataset.map, clears invalid
        parameters in the case where streaming is enabled

        :param dataset: dataset to apply mapping to
        :param kwargs: args to pass on to map function
        :return: mapped dataset
        """
        if isinstance(dataset, IterableDataset):
            # remove arguments that don't apply to streaming
            kwargs.pop("num_proc", None)
            kwargs.pop("load_from_cache_file", None)
            kwargs.pop("desc", None)

        return dataset.map(**kwargs)



import re
import weakref
from functools import wraps
from typing import Optional

import torch
import transformers
from accelerate.accelerator import get_state_dict_offloaded_model
from compressed_tensors import ModelCompressor, SparsityCompressionConfig
from loguru import logger
from transformers import PreTrainedModel

from llmcompressor.transformers.compression.quantization_format import (
    infer_quantization_format,
)
from llmcompressor.transformers.compression.sparsity_config import (
    SparsityConfigMetadata,
)

__all__ = ["modify_save_pretrained"]


def modify_save_pretrained(model: PreTrainedModel):
    """
    Overrides a PreTrainedModel's save_pretrained() method with a wrapped version that
    supports compression
    """

    def save_pretrained_compressed(save_pretrained_method):
        if getattr(save_pretrained_method, "_overridden", False):
            # `model.save_pretrained` has already been replaced, return.
            return save_pretrained_method

        # Keep a weak reference to the model class and unbound save_pretrained
        # method so we can call the original
        model_ref = weakref.ref(save_pretrained_method.__self__)
        original_save_pretrained = save_pretrained_method.__func__
        model_class = model_ref().__class__
        del save_pretrained_method

        @wraps(original_save_pretrained)
        def save_pretrained_wrapper(
            save_directory: str,
            sparsity_config: Optional[SparsityCompressionConfig] = None,
            quantization_format: Optional[str] = None,
            save_compressed: bool = True,
            skip_compression_stats: bool = False,
            **kwargs,
        ):
            """
            Wrapper around PreTrainedModel.save_pretrained(), adds functionality for
            saving models in a compressed format on disk. The compression format is
            saved to the model's config file

            :param save_directory: output directory to save model to
            :param sparsity_config: optional sparsity config to compress model with,
            if no config is provided it will be inferred from the model
            :param quantization_format: optional compression format for quantized
            models. If none is provided it will be inferred from the model
            :param save_compresed: whether or not to compress the model on disk
            :param skip_compression_stats: whether to skip the calculation of
            compression statistics (such as global sparsity and sparsity structure) when
            saving a model in dense format
            :param kwargs: additional kwargs to pass on to model.save_pretrained
            """

            # HACK: Override the dtype_byte_size function in transformers to
            # support float8 types. Fix is posted upstream
            # https://github.com/huggingface/transformers/pull/30488
            transformers.modeling_utils.dtype_byte_size = new_dtype_byte_size

            model = model_ref()
            # state_dict gets passed in as a kwarg for FSDP models
            state_dict = kwargs.get("state_dict", None)

            if sparsity_config is not None:
                sparsity_config.global_sparsity = (
                    SparsityConfigMetadata.infer_global_sparsity(
                        model, state_dict=state_dict
                    )
                )
                sparsity_config.sparsity_structure = (
                    SparsityConfigMetadata.infer_sparsity_structure()
                )
            elif not skip_compression_stats:
                # try to infer a sparsity config from the model if none is provided
                logger.info(
                    "Inferring a sparsity configuration requires a global sparsity "
                    "calculation. This can be costly for large models. To skip the "
                    "calculation of compression statistics set "
                    "skip_compression_stats=True"
                )
                sparsity_config = SparsityConfigMetadata.from_pretrained(
                    model, state_dict=state_dict, compress=False
                )

            quantization_format = infer_quantization_format(
                model=model,
                quantization_format=quantization_format,
                save_compressed=save_compressed,
                sparsity_config=sparsity_config,
            )
            compressor = ModelCompressor.from_pretrained_model(
                model,
                sparsity_config=sparsity_config,
                quantization_format=quantization_format,
            )

            if compressor is None:
                # model is not compressed or quantized, save as normal
                original_save_pretrained.__get__(model, model_class)(
                    save_directory, **kwargs
                )
                return

            # if we've gotten to this point we have a config so we can run compression
            # default safe serialization to True if not explicitly set
            kwargs["safe_serialization"] = kwargs.get("safe_serialization", True)
            if state_dict is None:
                state_dict = get_state_dict_offloaded_model(model)

            # make sure we're on the main process when saving
            if state_dict is not None and len(state_dict) > 0:
                compressed_state_dict = compressor.compress(model, state_dict)
                kwargs["state_dict"] = compressed_state_dict

                original_save_pretrained.__get__(model, model_class)(
                    save_directory, **kwargs
                )
                compressor.update_config(save_directory)

        save_pretrained_wrapper._overriden = True
        return save_pretrained_wrapper

    # wrap save_pretrained
    model.save_pretrained = save_pretrained_compressed(model.save_pretrained)


# HACK: Override the dtype_byte_size function in transformers to support float8 types
# Fix is posted upstream https://github.com/huggingface/transformers/pull/30488
def new_dtype_byte_size(dtype):
    if dtype == torch.bool:
        return 1 / 8
    bit_search = re.search(r"[^\d](\d+)_?", str(dtype))
    if bit_search is None:
        raise ValueError(f"`dtype` is not a valid dtype: {dtype}.")
    bit_size = int(bit_search.groups()[0])
    return bit_size // 8



from copy import deepcopy
from typing import Optional

from llmcompressor.transformers.finetune.data import TextGenerationDataset


@TextGenerationDataset.register(name="gsm8k")
class GSM8KDataset(TextGenerationDataset):
    """
    Child text generation class for the Grade School Math 8k dataset

    :param data_args: configuration settings for dataset loading
    :param split: split from dataset to load, for instance `test` or `train[:5%]`
    :param tokenizer: tokenizer to use on dataset
    """

    GSM_TEMPLATE = "Question: {question}\nAnswer:"

    def __init__(self, data_args, split, tokenizer):
        data_args = deepcopy(data_args)
        data_args.dataset = "gsm8k"
        super().__init__(
            text_column="text", data_args=data_args, split=split, tokenizer=tokenizer
        )

    def get_raw_dataset(self, cache_dir: Optional[str] = None):
        """
        Load the raw dataset from Hugging Face, using cached copy if available.
        Additionally reformats the entries to fit the alpaca template.

        :param cache_dir: disk location to search for cached dataset
        :return: the requested dataset
        """
        raw_dataset = super().get_raw_dataset(cache_dir=cache_dir)

        # helper fn for restructuring each dataset entry using the gsm template
        def restructure_fn(sample):
            sample["text"] = self.GSM_TEMPLATE.format(question=sample["question"])
            sample[self.PROMPT_KEY] = sample["text"]
            if "answer" in sample:
                sample["text"] += " " + sample["answer"]
            return sample

        raw_dataset = self.map(
            raw_dataset,
            function=restructure_fn,
            batched=False,
            remove_columns=["question", "answer"],
            num_proc=self.data_args.preprocessing_num_workers,
            load_from_cache_file=not self.data_args.overwrite_cache,
            desc="Restructuring GSM Dataset",
        )
        return raw_dataset



"""
Module for defining and managing optimizer event lifecycles in the LLM Compressor.

This module provides a class for defining event lifecycles when the optimizer is wrapped
to invoke the event lifecycle and no callbacks are used.
"""

from typing import List

from loguru import logger

from llmcompressor.core.events.event import Event, EventType
from llmcompressor.core.events.event_lifecycle import EventLifecycle

__all__ = [
    "OptimizerEventLifecycle",
]


class OptimizerEventLifecycle(EventLifecycle):
    """
    An event lifecycle for when the optimizer is wrapped to invoke the event lifecycle
    and no callbacks are used.

    For all flows with the OptimizerEventLifecycle, the optimizer is wrapped
    to trigger around the step function on the optimizer.
    loss_calculated is optional, but if used it must be consistently
    called before optimizer steps.

    The expected lifecycle is as follows with no gradient accumulation:
        for batch in training loop:
            if loss callbacks:
                loss_calculated() -> [BATCH_START, LOSS_CALCULATED]
                optim.step() -> [BATCH_START, OPTIM_PRE_STEP,
                                 OPTIM_POST_STEP, BATCH_END]
            else:
                optim.step() -> [BATCH_START, OPTIM_PRE_STEP,
                                 OPTIM_POST_STEP, BATCH_END]
    For gradient accumulation:
        for gradient_batches in training loop:
            for batch in gradient_batches:
                if loss callbacks:
                    if not last batch:
                        loss_calculated() -> [BATCH_START, LOSS_CALCULATED, BATCH_END]
                    else:
                        loss_calculated() -> [BATCH_START, LOSS_CALCULATED]
                        optim.step() -> [OPTIM_PRE_STEP, OPTIM_POST_STEP, BATCH_END]
                else:
                    if last batch:
                        optim.step() -> [BATCH_START, OPTIM_PRE_STEP,
                                         OPTIM_POST_STEP, BATCH_END]
    For older amp scale flows that use invocations_per_step > 1:
        for batch in training loop:
            for invocation in range(invocations_per_step):
                if not last invocation:
                    if loss callbacks:
                        loss_calculated() -> [BATCH_START, LOSS_CALCULATED]
                        optim.step() -> [BATCH_END]
                    else:
                        optim.step() -> [BATCH_START, BATCH_END]
                else:
                    if loss callbacks:
                        loss_calculated() -> [BATCH_START, LOSS_CALCULATED]
                        optim.step() -> [OPTIM_PRE_STEP, OPTIM_POST_STEP, BATCH_END]
                    else:
                        optim.step() -> [BATCH_START, OPTIM_PRE_STEP,
                                         OPTIM_POST_STEP, BATCH_END]

        - batch_start: must not be invoked, auto triggered
            from loss calculated if that is called, otherwise from pre_step
        - loss_calculated: optional pathway and invoked through callbacks.
            It must be called as the first event if used,
            and after optim post step or batch end for subsequent calls.
        - batch_end: must not be invoked, auto triggered from optim_post_step
        - optim_pre_step: must be called before optim_post_step
        - optim_post_step: must be called only once after optim_pre_step
    """

    def __init__(self, type_first: EventType, start: Event):
        """
        Initialize the OptimizerEventLifecycle with the first type and start event.

        :param type_first: The first event type to be called
        :type type_first: EventType
        :param start: The start event to base the lifecycle off of
        :type start: Event
        """
        super().__init__(type_first=type_first, start=start)
        self.skip_post_step = False

    def batch_start_events(self) -> List[Event]:
        """
        Raises a ValueError as this method should not be called.

        :raises ValueError: If invoked as this should not be called
        """
        logger.error("batch start should not be invoked when only wrapped optim")
        raise ValueError("batch start should not be invoked when only wrapped optim")

    def loss_calculated_events(self) -> List[Event]:
        """
        Return the list of events to be called for the loss calculated.

        :return: The list of events to be called for the loss calculated
        :rtype: List[Event]
        :raises ValueError: If invoked before loss calculation
        """
        if self.type_first != EventType.LOSS_CALCULATED:
            logger.error("loss calculated must be called first for wrapped optim")
            raise ValueError("loss calculated must be called first for wrapped optim")

        if self.last_type not in {
            EventType.LOSS_CALCULATED,
            EventType.OPTIM_POST_STEP,
            None,
        }:
            logger.error(
                "loss calculated must be called after batch end or optim post step"
            )
            raise ValueError(
                "loss calculated must be called after batch end or optim post step"
            )

        self.last_type = EventType.LOSS_CALCULATED

        if not self.check_batches_per_step_count(increment=True):
            logger.debug(
                "Loss calculated event processed, "
                "step not ready at global_step: {} and global_batch: {}",
                self.global_step,
                self.global_batch,
            )
            return [
                self.new_instance(type_=EventType.BATCH_START),
                self.new_instance(type_=EventType.LOSS_CALCULATED),
                self.new_instance(type_=EventType.BATCH_END),
            ]
        else:
            logger.debug(
                "Loss calculated event processed, "
                "step ready at global_step: {} and global_batch: {}",
                self.global_step,
                self.global_batch,
            )
            return [
                self.new_instance(type_=EventType.BATCH_START),
                self.new_instance(type_=EventType.LOSS_CALCULATED),
            ]

    def optim_pre_step_events(self) -> List[Event]:
        """
        Return the list of events to be called for the optim pre step.

        :return: The list of events to be called for the optim pre step
        :rtype: List[Event]
        :raises ValueError: If optim pre step is not called before optim post step
            or after loss calculated
        """
        if self.type_first == EventType.LOSS_CALCULATED:
            # handle loss calculated case where gradient accumulation
            # is automatically handled by the loss callbacks

            if self.last_type != EventType.LOSS_CALCULATED:
                logger.error("optim pre step must be called after loss calculated")
                raise ValueError("optim pre step must be called after loss calculated")

            self.last_type = EventType.OPTIM_PRE_STEP

            if not self.check_invocations_per_step_count(increment=False):
                logger.debug(
                    "Optim pre step event processed, "
                    "but invocations not ready at global_step: {} and global_batch: {}",
                    self.global_step,
                    self.global_batch,
                )
                self.skip_post_step = True
                return []
            else:
                logger.debug(
                    "Optim pre step event processed, "
                    "invocations ready at global_step: {} and global_batch: {}",
                    self.global_step,
                    self.global_batch,
                )
                self.skip_post_step = False
                return [self.new_instance(type_=EventType.OPTIM_PRE_STEP)]

        # handle no callbacks case to emulate batch events for gradient accumulation
        if self.last_type not in {EventType.OPTIM_POST_STEP, None}:
            logger.error(
                "optim pre step must be called at the start or after optim post step"
            )
            raise ValueError(
                "optim pre step must be called at the start or after optim post step"
            )

        batch_events = [
            self.new_instance(type_=EventType.BATCH_START),
        ]
        while not self.check_batches_per_step_count(increment=True):
            batch_events.append(self.new_instance(type_=EventType.BATCH_END))
            batch_events.append(self.new_instance(type_=EventType.BATCH_START))

        self.last_type = EventType.OPTIM_PRE_STEP

        if not self.check_invocations_per_step_count(increment=False):
            logger.debug(
                "Optim pre step event processed, "
                "but invocations not ready at global_step: {} and global_batch: {}",
                self.global_step,
                self.global_batch,
            )
            self.skip_post_step = True
            return batch_events
        else:
            logger.debug(
                "Optim pre step event processed, "
                "invocations ready at global_step: {} and global_batch: {}",
                self.global_step,
                self.global_batch,
            )
            self.skip_post_step = False
            return batch_events + [self.new_instance(type_=EventType.OPTIM_PRE_STEP)]

    def optim_post_step_events(self) -> List[Event]:
        """
        Return the list of events to be called for the optim post step.

        :return: The list of events to be called for the optim post step
        :rtype: List[Event]
        :raises ValueError: If optim post step is not called after optim pre step
        """
        if self.last_type != EventType.OPTIM_PRE_STEP:
            logger.error("optim post step must be called after optim pre step")
            raise ValueError("optim post step must be called after optim pre step")

        self.last_type = EventType.OPTIM_POST_STEP

        if self.skip_post_step:
            logger.debug(
                "Skipping optim post step event at global_step: "
                "{} and global_batch: {}",
                self.global_step,
                self.global_batch,
            )
            return [self.new_instance(type_=EventType.BATCH_END)]
        else:
            logger.debug(
                "Optim post step event processed at global_step: "
                "{} and global_batch: {}",
                self.global_step,
                self.global_batch,
            )
            return [
                self.new_instance(type_=EventType.OPTIM_POST_STEP),
                self.new_instance(type_=EventType.BATCH_END),
            ]

    def batch_end_events(self) -> List[Event]:
        """
        Raises a ValueError as this method should not be called.

        :raises ValueError: If invoked as this should not be called
        """
        logger.error("batch end should not be invoked when only wrapped optim")
        raise ValueError("batch end should not be invoked when only wrapped optim")



# Copyright (c) 2021 - present / Neuralmagic, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from copy import deepcopy
from typing import Dict, List, Union

from datasets.dataset_dict import Dataset, DatasetDict

from llmcompressor.transformers.finetune.data import TextGenerationDataset
from llmcompressor.transformers.utils.preprocessing_functions import (
    PreprocessingFunctionRegistry,
)
from llmcompressor.utils import import_from_path


@TextGenerationDataset.register(name="custom", alias=["json", "csv"])
class CustomDataset(TextGenerationDataset):
    """
    Child text generation class for custom local dataset supporting load
    for csv and json

    :param data_args: configuration settings for dataset loading
    :param split: split from dataset to load, for instance `test` or `train[:5%]`
        Can also be set to None to load all the splits
    :param tokenizer: tokenizer to use on dataset

    """

    def __init__(self, data_args, split, tokenizer):
        data_args = deepcopy(data_args)
        super().__init__(
            text_column=data_args.text_column,
            data_args=data_args,
            split=split,
            tokenizer=tokenizer,
        )
        self.preprocessing_func = data_args.preprocessing_func
        self.remove_columns = data_args.remove_columns

    def get_raw_dataset(self, *_ignore, **__ignore) -> Union[DatasetDict, Dataset]:
        """Get the raw dataset and apply preprocessing func if provided"""

        dataset = self.data_args.dataset
        if isinstance(dataset, DatasetDict) or isinstance(dataset, Dataset):
            # user passed in an already instantiated dataset, just use it directly
            raw_dataset = dataset
        else:
            # dataset must be loaded from file or HF Hub
            raw_dataset = super().get_raw_dataset()

        if self.preprocessing_func is not None:
            if callable(self.preprocessing_func):
                func = self.preprocessing_func
            elif ":" in self.preprocessing_func:
                # load func_name from "/path/to/file.py:func_name"
                func = import_from_path(self.preprocessing_func)
            else:
                # load from the registry
                func = PreprocessingFunctionRegistry.get_value_from_registry(
                    name=self.preprocessing_func
                )

            raw_dataset = self.map(
                raw_dataset,
                function=func,
                batched=False,
                num_proc=self.data_args.preprocessing_num_workers,
                desc="Applying custom func to the custom dataset",
            )

        self.remove_columns = (
            self.remove_columns or self.get_remove_columns_from_dataset(raw_dataset)
        )

        if self.remove_columns is not None:
            raw_dataset = self.map(
                raw_dataset,
                batched=True,
                remove_columns=self.remove_columns,
                num_proc=self.data_args.preprocessing_num_workers,
                desc="Removing unneeded columns",
            )

        return raw_dataset

    def get_remove_columns_from_dataset(
        self, raw_dataset: Union[DatasetDict, Dataset]
    ) -> List[str]:
        """Remove redandant columns from the dataset for processing"""

        remove_columns = raw_dataset.column_names
        if isinstance(remove_columns, Dict):
            remove_columns = raw_dataset[list(raw_dataset.keys())[0]].column_names

        remove_columns = set(remove_columns)
        if self.text_column in remove_columns:
            remove_columns.remove(self.text_column)
        if self.PROMPT_KEY in remove_columns:
            remove_columns.remove(self.PROMPT_KEY)

        return list(remove_columns)



import operator
from pathlib import Path
from typing import Optional

from loguru import logger

try:
    from torch.distributed.fsdp import (
        FullStateDictConfig,
        FullyShardedDataParallel,
        StateDictType,
    )
except ImportError:
    FullyShardedDataParallel = None

import torch
from torch.nn import Module

from llmcompressor.core.state import State
from llmcompressor.pytorch.model_load.helpers import save_model_and_recipe
from llmcompressor.utils.pytorch import set_layer

__all__ = [
    "is_fsdp_model",
    "maybe_get_wrapped",
    "set_wrapped_model",
    "unwrap_and_export_model",
    "save_pretrained_fsdp",
    "get_fsdp_parent",
    "find_and_move_state_dicts_to_cpu",
]


def is_fsdp_model(model: Module) -> bool:
    """
    Check if a model instance is wrapped by FSDP

    :param model: pytorch model to check
    :return: True if module is wrapped, False otherwise
    """
    if not FullyShardedDataParallel:
        return False

    return isinstance(model, FullyShardedDataParallel)


def maybe_get_wrapped(model: Module) -> Module:
    """
    Given a model that may or may not have a distributed wrapper, return the underlying
    wrapped model.

    :param model: input model to get wrapped model from
    :returns: wrapped model
    """
    if is_fsdp_model(model=model):
        return model._fsdp_wrapped_module
    return model


def set_wrapped_model(state: State, wrapped_model: Module):
    """
    Given a state with a model that may or may not have a distributed wrapper, set
    the underlying wrapped model.

    :param state: state to update model of
    :param updated_wrapped: model to inject into input_model
    """
    if is_fsdp_model(state.model):
        state.model._fsdp_wrapped_module = wrapped_model
    else:
        state.model = wrapped_model


def unwrap_and_export_model(model, accelerator, output_dir, tokenizer):
    """
    Recursively unwraps an FSDP model, then saves the unwrapped model and the
    currently active recipe to disk

    :param model: model to unwrap
    :param accelerator: Accelerator instance used to perform unwrapping
    :param output_dir: where to save output model
    :param tokenizer: tokenizer used by the model
    """
    full_state_dict_config = FullStateDictConfig(offload_to_cpu=True, rank0_only=True)
    with FullyShardedDataParallel.state_dict_type(
        model,
        StateDictType.FULL_STATE_DICT,
        full_state_dict_config,
    ):
        unwrapped_model = accelerator.unwrap_model(model)
        for name, module in unwrapped_model.named_modules():
            if isinstance(module, FullyShardedDataParallel):
                set_layer(name, accelerator.unwrap_model(module), unwrapped_model)

        save_model_and_recipe(
            model=unwrapped_model,
            save_path=output_dir,
            tokenizer=tokenizer,
        )


def find_and_move_state_dicts_to_cpu(output_dir: str):
    """
    Looks for state dicts in the output directory and overwrites them
    with cpu state dicts.

    this is needed for quantized models trained with FSDP as the state dict
    contains device information, which can cause issues when loading the model
    using transformers AutoModel.from_pretrained(...) if the device information
    is not removed, assumes the state dicts are named pytorch_model*.bin
    """

    for model_file in Path(output_dir).rglob("pytorch_model*.bin"):
        loaded_dict = torch.load(model_file)
        for key, value in loaded_dict.items():
            if isinstance(value, torch.Tensor):
                loaded_dict[key] = value.cpu()

        torch.save(loaded_dict, model_file)
        logger.info(f"Moved state dict {model_file} to cpu")


def save_pretrained_fsdp(
    model,
    accelerator,
    output_dir,
    save_safetensors: bool = True,
    save_compressed: bool = False,
):
    full_state_dict_config = FullStateDictConfig(offload_to_cpu=True, rank0_only=True)
    """
    Gathers the full FSDP state dict of the model onto rank0 GPU, then uses it to save
    the pretrained FSDP model to disk

    :param model: model to save
    :param accelerator: Accelerator instance used to perform unwrapping
    :param output_dir: where to save output model
    :param save_safetensors: True to safe in safetensors format, otherwise .bin
    :param save_compressed: whether to compress sparse weights on disk
    """
    with FullyShardedDataParallel.state_dict_type(
        model, StateDictType.FULL_STATE_DICT, full_state_dict_config
    ):
        state_dict = accelerator.get_state_dict(model, unwrap=False)

    if accelerator.is_main_process:
        accelerator.unwrap_model(model).save_pretrained(
            output_dir,
            is_main_process=accelerator.is_main_process,
            save_function=accelerator.save,
            state_dict=state_dict,
            save_compressed=save_compressed,
            safe_serialization=save_safetensors,
        )

    accelerator.wait_for_everyone()


def get_fsdp_parent(layer_name: str, model: Module) -> Optional[Module]:
    """
    Gets the closest parent of layer_name that is wrapped by FSDP. If no FSDP wrapper
    is found just return None

    :param layer_name: layer name in model to get parent of
    :model: pytorch module to search through
    :return: FSDP wrapped parent of layer_name if available, otherwise None
    """
    if not is_fsdp_model(model):
        return None

    parent_name = layer_name
    parent = operator.attrgetter(parent_name)(model)
    while not isinstance(parent, FullyShardedDataParallel):
        if len(parent_name) == 0:  # we've reached the root module and its not FSDP
            # this should never get hit because we check for an FSDP root above
            # but while statements without a backup are too scary
            return None
        parent_name = ".".join(parent_name.split(".")[:-1])
        parent = operator.attrgetter(parent_name)(model)

    return parent



import operator
from collections import defaultdict
from math import ceil
from typing import List, Optional

import torch
from loguru import logger
from torch.nn.modules.sparse import Embedding

_DEFAULT_TARGET_IDS = ["attention_mask", "position_ids", "position_bias"]


class Catcher(torch.nn.Module):
    def __init__(self, module, target_keys: Optional[List[str]] = None):
        super().__init__()
        self.module = module
        self.target_keys = target_keys
        self.cache = defaultdict(list)
        self.cache["inputs"] = []

    def forward(self, *args, **kwargs):
        self.cache["inputs"].append(args)
        if self.target_keys is None:
            self.target_keys = self._get_target_keys(kwargs.keys())

        for key in self.target_keys:
            self.cache[key].append(kwargs[key])
        raise ValueError

    def get_cache(self):
        return self.cache

    def _get_target_keys(self, input_keys):
        target_keys = []
        for key in _DEFAULT_TARGET_IDS:
            if key in input_keys:
                target_keys.append(key)
        return target_keys


def replace_module(model, old_module, new_module):
    for module_name, module in model.named_modules():
        if old_module == module:
            break

    current_module = model
    module_name = module_name.split(".")
    for child_module in module_name[:-1]:
        current_module = getattr(current_module, child_module)
    setattr(current_module, module_name[-1], new_module)


def catch(model, attention_layer, target_keys, data_loader, nsamples):
    catcher_module = Catcher(attention_layer, target_keys)
    replace_module(model, attention_layer, catcher_module)
    device = next(attention_layer.parameters()).device
    for input_id, inp in enumerate(data_loader):
        if nsamples is not None and input_id == nsamples:
            break
        try:
            if isinstance(inp, tuple):
                inp = inp[0]
            model(inp.to(device), use_cache=False)
        except ValueError:
            pass
    replace_module(model, catcher_module, attention_layer)
    return catcher_module.get_cache()


def execute_offloaded_module(
    module,
    buffer,
    dev,
    nsamples=None,
    overwrite_buffer=True,
    cached_inputs=None,
    **kwargs,
):
    module.to(dev)
    if not overwrite_buffer:
        new_buffer = []
    for input_index, inp in enumerate(buffer):
        if nsamples is not None and input_index == nsamples:
            break
        if cached_inputs is None:
            module_kwargs = kwargs
        else:
            module_kwargs = {
                key: cached_inputs[key][input_index] for key in cached_inputs
            }
            module_kwargs.update(kwargs)
        if isinstance(inp, tuple):
            inp = inp[0]
        output = module(inp.to(dev), **module_kwargs)
        if overwrite_buffer:
            buffer[input_index] = output
        else:
            new_buffer.append(output)

    module.cpu()
    torch.cuda.empty_cache()
    if overwrite_buffer:
        return buffer
    else:
        del buffer
        torch.cuda.empty_cache()
        return new_buffer


def cache_attention_inputs(
    model, dataloader, device, nsamples, target_ids, layer_prefix
):
    if layer_prefix:  # get model-specific path to layers list
        split_prefix = layer_prefix.split(".")
        layers_name = split_prefix[-1]
        model_root_name = ".".join(split_prefix[:-1])
        model_root = operator.attrgetter(model_root_name)(model)
        first_layer = getattr(model_root, layers_name)[0]
    else:
        model_root = model.model
        layers_name = "layers"
        first_layer = model_root.layers[0]

    # send everything up to the first compressable layer to device
    pre_layers_modules = _get_pre_layer_modules(model_root, layers_name)
    for pre_layer in pre_layers_modules:
        pre_layer.to(device)
    first_layer.to(device)

    cached_inputs = catch(
        model=model,
        attention_layer=first_layer,
        target_keys=target_ids,
        data_loader=dataloader,
        nsamples=nsamples,
    )

    for pre_layer in pre_layers_modules:
        pre_layer.cpu()
    first_layer.cpu()
    torch.cuda.empty_cache()

    return cached_inputs


@torch.no_grad()
def ppl_eval_general(
    eval_logits, model, dataloader, dev, nsamples=None, max_samples_per_iteration=128
):
    logger.info("Evaluating perplexity...")

    if nsamples is None:
        nsamples = len(dataloader)

    number_iterations = int(ceil(nsamples / max_samples_per_iteration))
    neg_log_likelihood = 0.0
    number_tokens = 0
    for iteration in range(number_iterations):
        if iteration < number_iterations - 1:
            samples = dataloader[
                iteration * max_samples_per_iteration : (iteration + 1)
                * max_samples_per_iteration
            ]
        else:
            samples = dataloader[iteration * max_samples_per_iteration :]

        logits = eval_logits(model, samples, dev)

        vocabulary_size = logits[0].shape[-1]
        logits = [logit[:, :-1, :].view(-1, vocabulary_size) for logit in logits]
        logits = torch.cat(logits, dim=0).contiguous().to(torch.float32)

        labels = [sample[:, 1:].view(-1) for sample in samples]
        labels = torch.cat(labels, dim=0).to(dev)
        neg_log_likelihood += torch.nn.functional.cross_entropy(
            logits,
            labels,
            reduction="sum",
        )

        number_tokens += labels.numel()
        perplexity = torch.exp(neg_log_likelihood / number_tokens)
        logger.debug(
            f"Processed iteration {iteration} of {number_iterations} with perplexity: "
            f"{perplexity}"
        )

    perplexity = torch.exp(neg_log_likelihood / number_tokens)
    logger.info(f"Evaluated perplexity: {perplexity}")

    return perplexity.item()


def _get_pre_layer_modules(model_root, layers_name):
    pre_layers_modules = []
    for name, layer in model_root.named_modules():
        if name.startswith(layers_name):
            break
        if isinstance(layer, Embedding):
            pre_layers_modules.append(layer)

    return pre_layers_modules



"""
Helper variables and functions for integrating LLM Compressor with
huggingface/transformers flows
"""

import inspect
import os
from collections import OrderedDict
from contextlib import suppress
from enum import Enum
from pathlib import Path
from typing import Iterable, List, Optional
from typing import OrderedDict as OrderedDictType
from typing import Tuple, Union

import requests
import torch
import transformers
from huggingface_hub import HUGGINGFACE_CO_URL_HOME, HfFileSystem, hf_hub_download
from loguru import logger
from transformers import AutoConfig
from transformers.trainer_utils import get_last_checkpoint
from transformers.utils import PaddingStrategy

from llmcompressor.utils.fsdp.context import main_process_first_context

__all__ = [
    "RECIPE_NAME",
    "detect_last_checkpoint",
    "TaskNames",
    "resolve_sequence_length",
    "ALL_TASK_NAMES",
    "create_fake_dataloader",
    "POSSIBLE_TOKENIZER_FILES",
    "download_repo_from_huggingface_hub",
    "download_model_directory",
]


class TaskNames(Enum):
    mlm = {"masked-language-modeling", "mlm"}
    qa = {"question-answering", "qa"}
    token_classification = {"token-classification", "ner"}
    text_classification = {
        "text-classification",
        "sentiment-analysis",
        "sequence-classification",
        "glue",
    }
    text_generation = {"text-generation"}


ALL_TASK_NAMES = list(set.union(*[task_names.value for task_names in TaskNames]))
RECIPE_NAME = "recipe.yaml"

MANDATORY_DEPLOYMENT_FILES = {
    "tokenizer_config.json",
    "config.json",
}
OPTIONAL_DEPLOYMENT_FILES = {"tokenizer.json", "tokenizer.model"}
NLG_MANDATORY_DEPLOYMENT_FILES = {"special_tokens_map.json"}
NLG_OPTIONAL_DEPLOYMENT_FILES = {
    "vocab.json",
    "merges.txt",
}
POSSIBLE_TOKENIZER_FILES = {
    "vocab.json",
    "merges.txt",
    "tokenizer.json",
    "tokenizer.model",
    "special_tokens_map.json",
    "tokenizer_config.json",
}
RELEVANT_HF_SUFFIXES = ["json", "md", "bin", "safetensors", "yaml", "yml", "py"]


def detect_last_checkpoint(
    training_args: "TrainingArguments",  # noqa 821
    model_args: Optional["ModelArguments"] = None,  # noqa 821
):
    last_checkpoint = None
    if (
        os.path.isdir(training_args.output_dir)
        and training_args.do_train
        and not training_args.overwrite_output_dir
    ):
        last_checkpoint = get_last_checkpoint(training_args.output_dir)
        if training_args.run_stages and model_args is not None:
            model = (
                model_args.model
                if hasattr(model_args, "model")
                else model_args.model_name_or_path
            )
            if os.path.isdir(model):
                last_checkpoint = get_last_checkpoint(model_args.model_name_or_path)
        if last_checkpoint is None and (len(os.listdir(training_args.output_dir)) > 0):
            raise ValueError(
                f"Output directory ({training_args.output_dir}) already "
                "exists and is not empty. Use --overwrite_output_dir to overcome."
            )
        elif (
            last_checkpoint is not None and training_args.resume_from_checkpoint is None
        ):
            logger.info(
                f"Checkpoint detected, resuming training at {last_checkpoint}. To "
                "avoid this behavior, change  the `--output_dir` or add "
                "`--overwrite_output_dir` to train from scratch."
            )

    return last_checkpoint


def resolve_sequence_length(config: AutoConfig) -> int:
    """
    Resolve the sequence length from the config

    :param config: the config to resolve the sequence length from
    :return: the sequence length
    """
    if hasattr(config, "max_position_embeddings"):
        sequence_length = config.max_position_embeddings

    elif hasattr(config, "max_seq_len"):
        sequence_length = config.max_seq_len
    else:
        raise ValueError(
            "Could not infer a default sequence length "
            "from the HF transformers config. Please specify "
            "the sequence length with --sequence_length"
        )
    logger.debug(
        f"Using default sequence length of {sequence_length} "
        "(inferred from HF transformers config) "
    )
    return sequence_length


def resolve_recipe(
    model_path: Union[str, Path],
    recipe: Union[str, Path, None] = None,
) -> Union[str, None]:
    """
    Resolve the recipe to apply to the model.
    :param recipe: the recipe to apply to the model.
        It can be one of the following:
        - None
            This means that we are not either not applying
            any recipe and allowing the model to potentially
            infer the appropriate pre-existing recipe
            from the model_path
        - a path to the recipe file
            This can be a string or Path object pointing
            to a recipe file. If the specified recipe file
            is different from the potential pre-existing
            recipe for that model (stored in the model_path),
            the function will raise an warning
        - name of the recipe file (e.g. "recipe.yaml")
            Recipe file name specific is assumed to be stored
            in the model_path
        - a string containing the recipe
            Needs to adhere to the SparseML recipe format

    :param model_path: the path to the model to load.
        It can be one of the following:
        - a path to the model directory
        - a path to the model file
        - Hugging face model id

    :return: the resolved recipe
    """

    if recipe is None:
        return infer_recipe_from_model_path(model_path)

    elif os.path.isfile(recipe):
        # recipe is a path to a recipe file
        return resolve_recipe_file(recipe, model_path)

    elif os.path.isfile(os.path.join(model_path, recipe)):
        # recipe is a name of a recipe file
        recipe = os.path.join(model_path, recipe)
        return resolve_recipe_file(recipe, model_path)

    elif isinstance(recipe, str):
        # recipe is a string containing the recipe
        logger.debug(
            "Applying the recipe string directly to the model, without "
            "checking for a potential existing recipe in the model_path."
        )
        return recipe

    logger.info(
        "No recipe requested and no default recipe "
        f"found in {model_path}. Skipping recipe resolution."
    )
    return None


def infer_recipe_from_model_path(model_path: Union[str, Path]) -> Optional[str]:
    """
    Infer the recipe from the model_path.
    :param model_path: the path to the model to load.
        It can be one of the following:
        - a path to the model directory
        - a path to the model file
        - Hugging face model id
    :return the path to the recipe file if found, None otherwise
    """
    model_path = model_path.as_posix() if isinstance(model_path, Path) else model_path

    if os.path.isdir(model_path) or os.path.isfile(model_path):
        # model_path is a local path to the model directory or model file
        # attempting to find the recipe in the model_directory
        model_path = (
            os.path.dirname(model_path) if os.path.isfile(model_path) else model_path
        )
        recipe = os.path.join(model_path, RECIPE_NAME)
        if os.path.isfile(recipe):
            logger.info(f"Found recipe in the model_path: {recipe}")
            return recipe
        logger.debug(f"No recipe found in the model_path: {model_path}")
        return None

    recipe = recipe_from_huggingface_model_id(model_path)[0]

    if recipe is None:
        logger.info("Failed to infer the recipe from the model_path")
    return recipe


def recipe_from_huggingface_model_id(
    model_path: str, recipe_name: str = RECIPE_NAME
) -> Tuple[Optional[str], bool]:
    """
    Attempts to download the recipe from the huggingface model id.

    :param model_path: Assumed to be the huggingface model id.
        If it is not, this function will return None.
    :param recipe_name: The name of the recipe file to download.
        Defaults to RECIPE_NAME.
    :return: tuple:
        - the path to the recipe file if found, None otherwise
        - True if model_path is a valid huggingface model id, False otherwise
    """
    model_id = os.path.join(HUGGINGFACE_CO_URL_HOME, model_path)
    request = requests.get(model_id)
    if not request.status_code == 200:
        logger.debug(
            "model_path is not a valid huggingface model id. "
            "Skipping recipe resolution."
        )
        return None, False

    logger.info(
        "model_path is a huggingface model id. "
        "Attempting to download recipe from "
        f"{HUGGINGFACE_CO_URL_HOME}"
    )
    try:
        recipe = hf_hub_download(repo_id=model_path, filename=recipe_name)
        logger.info(f"Found recipe: {recipe_name} for model id: {model_path}.")
    except Exception as e:
        logger.info(
            f"Unable to to find recipe {recipe_name} "
            f"for model id: {model_path}: {e}. "
            "Skipping recipe resolution."
        )
        recipe = None
    return recipe, True


def resolve_recipe_file(
    requested_recipe: Union[str, Path], model_path: Union[str, Path]
) -> Union[str, Path, None]:
    """
    Given the requested recipe and the model_path, return the path to the recipe file.

    :param requested_recipe. Is a full path to the recipe file
    :param model_path: the path to the model to load.
        It can be one of the following:
        - a path to the model directory
        - a path to the model file
        - Hugging face model id
    :return the path to the recipe file if found, None otherwise
    """
    # preprocess arguments so that they are all strings
    requested_recipe = (
        requested_recipe.as_posix()
        if isinstance(requested_recipe, Path)
        else requested_recipe
    )
    model_path = model_path.as_posix() if isinstance(model_path, Path) else model_path
    model_path = (
        os.path.dirname(model_path) if os.path.isfile(model_path) else model_path
    )

    if not os.path.isdir(model_path):
        default_recipe, model_exists = recipe_from_huggingface_model_id(model_path)
        if not model_exists:
            raise ValueError(f"Unrecognized model_path: {model_path}")

        if not default_recipe == requested_recipe and default_recipe is not None:
            logger.warning(
                f"Attempting to apply recipe: {requested_recipe} "
                f"to the model at: {model_path}, "
                f"but the model already has a recipe: {default_recipe}. "
                f"Using {requested_recipe} instead."
            )
        return requested_recipe

    # pathway for model_path that is a directory
    default_recipe = os.path.join(model_path, RECIPE_NAME)
    default_recipe_exists = os.path.isfile(default_recipe)
    default_and_request_recipes_identical = os.path.samefile(
        default_recipe, requested_recipe
    )

    if (
        default_recipe_exists
        and requested_recipe
        and not default_and_request_recipes_identical
    ):
        logger.warning(
            f"Attempting to apply recipe: {requested_recipe} "
            f"to the model located in {model_path}, "
            f"but the model already has a recipe stored as {default_recipe}. "
            f"Using {requested_recipe} instead."
        )

    elif not default_recipe_exists and requested_recipe:
        logger.warning(
            f"Attempting to apply {requested_recipe} "
            f"to the model located in {model_path}."
            "However, it is expected that the model "
            f"has its target recipe stored as {default_recipe}."
            "Applying any recipe before the target recipe may "
            "result in unexpected behavior."
            f"Applying {requested_recipe} nevertheless."
        )

    elif default_recipe_exists:
        logger.info(f"Using the default recipe: {requested_recipe}")

    return requested_recipe


def create_fake_dataloader(
    model: torch.nn.Module,
    tokenizer: transformers.AutoTokenizer,
    num_samples: int,
) -> Tuple[Iterable[OrderedDictType[str, torch.Tensor]], List[str]]:
    """
    Creates fake transformers dataloader for the model, based on the model's
    forward signature.

    :param model: The model to create the dataloader for
    :param tokenizer: The tokenizer to use for the dataloader
    :param num_samples: The number of fake samples in the dataloader
    :return: The data loader (iterable) and the input names for the model
    """

    forward_args_spec = inspect.getfullargspec(model.__class__.forward)
    inputs = tokenizer(
        "", return_tensors="pt", padding=PaddingStrategy.MAX_LENGTH.value
    ).data
    fake_inputs = OrderedDict(
        [
            (input_key, inputs[input_key][0].reshape(1, -1))
            for input_key in forward_args_spec.args
            if input_key in inputs
        ]
    )
    data_loader = (fake_inputs for _ in range(num_samples))
    input_names = list(fake_inputs.keys())
    return data_loader, input_names


def fetch_recipe_path(target: str):
    """
    Fetches the recipe path for the given target.
    This method will also download the recipe if it is not
    already downloaded.

    Takes care of three scenarios:
    1. target is a local path to a model directory
        (looks for recipe.yaml in the directory)
    2. target is a HuggingFace stub (downloads and
        returns the path to the default recipe)

    :param target: The target to fetch the recipe path for
        can be a local path or HuggingFace stub
    :return: The path to the recipe for the target
    """
    DEFAULT_RECIPE_NAME = "recipe.yaml"
    if Path(target).exists():
        # target is a local path
        potential_recipe_path = Path(target) / DEFAULT_RECIPE_NAME
        return str(potential_recipe_path) if potential_recipe_path.exists() else None

    # Recipe must be downloaded

    recipe_path = None

    # target is a HuggingFace stub
    with suppress(Exception):
        # suppress any errors if the recipe is not found on HuggingFace
        recipe_path = hf_hub_download(repo_id=target, filename=DEFAULT_RECIPE_NAME)

    return recipe_path


def download_repo_from_huggingface_hub(repo_id, **kwargs):
    """
    Download relevant model files from the Hugging Face Hub
    using the huggingface_hub.hf_hub_download function

    Note(s):
    - Does not download the entire repo, only the relevant files
    for the model, such as the model weights, tokenizer files, etc.
    - Does not re-download files that already exist locally, unless
    the force_download flag is set to True

    :pre-condition: the repo_id must be a valid Hugging Face Hub repo id
    :param repo_id: the repo id to download
    :param kwargs: additional keyword arguments to pass to hf_hub_download
    """
    hf_filesystem = HfFileSystem()
    files = hf_filesystem.ls(repo_id)

    if not files:
        raise ValueError(f"Could not find any files in HF repo {repo_id}")

    # All file(s) from hf_filesystem have "name" key
    # Extract the file names from the files
    relevant_file_names = (
        Path(file["name"]).name
        for file in files
        if any(file["name"].endswith(suffix) for suffix in RELEVANT_HF_SUFFIXES)
    )

    hub_kwargs_names = (
        "subfolder",
        "repo_type",
        "revision",
        "library_name",
        "library_version",
        "cache_dir",
        "local_dir",
        "local_dir_use_symlinks",
        "user_agent",
        "force_download",
        "force_filename",
        "proxies",
        "etag_timeout",
        "resume_download",
        "token",
        "local_files_only",
        "headers",
        "legacy_cache_layout",
        "endpoint",
    )
    hub_kwargs = {name: kwargs[name] for name in hub_kwargs_names if name in kwargs}

    for file_name in relevant_file_names:
        last_file = hf_hub_download(repo_id=repo_id, filename=file_name, **hub_kwargs)

    # parent directory of the last file is the model directory
    return str(Path(last_file).parent.resolve().absolute())


def download_model_directory(pretrained_model_name_or_path: str, **kwargs):
    """
    Download the model directory from the HF hub if the model is not found locally

    :param pretrained_model_name_or_path: the name of or path to the model to load
        can be a HuggingFace model stub
    :param kwargs: additional keyword arguments to pass to the download function
    :return: the path to the downloaded model directory
    """
    pretrained_model_path: Path = Path(pretrained_model_name_or_path)

    if pretrained_model_path.exists():
        logger.debug(
            "Model directory already exists locally.",
        )
        return pretrained_model_name_or_path

    with main_process_first_context():
        logger.debug("Downloading model from HuggingFace Hub.")
        return download_repo_from_huggingface_hub(
            repo_id=pretrained_model_name_or_path, **kwargs
        )



# Copyright (c) 2021 - present / Neuralmagic, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from copy import deepcopy
from typing import Optional

from llmcompressor.transformers.finetune.data import TextGenerationDataset


@TextGenerationDataset.register(name="open_platypus")
class OpenPlatypusDataset(TextGenerationDataset):
    """
    Child text generation class for the Open Platypus dataset

    :param data_args: configuration settings for dataset loading
    :param split: split from dataset to load, for instance `test` or `train[:5%]`
    :param tokenizer: tokenizer to use on dataset
    """

    ALPACA_TEMPLATE = {
        "prompt_input": "Below is an instruction that describes a task, paired with an "
        "input that provides further context. Write a response that appropriately "
        "completes the request.\n\n### Instruction:\n{instruction}\n\n### Input:\n"
        "{input}\n\n### Response:\n",
        "prompt_no_input": "Below is an instruction that describes a task. Write a "
        "response that appropriately completes the request.\n\n### Instruction:\n{"
        "instruction}\n\n### Response:\n",
    }

    def __init__(self, data_args, split, tokenizer):
        data_args = deepcopy(data_args)
        data_args.dataset = "garage-bAInd/Open-Platypus"
        super().__init__(
            text_column="text", data_args=data_args, split=split, tokenizer=tokenizer
        )

    def get_raw_dataset(self, cache_dir: Optional[str] = None):
        """
        Load the raw dataset from Hugging Face, using cached copy if available.
        Additionally reformats the entries to fit the alpaca template.

        :param cache_dir: disk location to search for cached dataset
        :return: the requested dataset
        """
        raw_dataset = super().get_raw_dataset(cache_dir=cache_dir)

        # helper fn for restructuring each dataset entry using the alpaca template
        def restructure_fn(sample):
            if "input" in sample and sample["input"] != "":
                sample["text"] = self.ALPACA_TEMPLATE["prompt_input"].format(
                    instruction=sample["instruction"], input=sample["input"]
                )
            else:
                sample["text"] = self.ALPACA_TEMPLATE["prompt_no_input"].format(
                    instruction=sample["instruction"]
                )

            sample[self.PROMPT_KEY] = sample["text"]
            if "output" in sample:
                sample["text"] += sample["output"]
            return sample

        raw_dataset = self.map(
            raw_dataset,
            function=restructure_fn,
            batched=False,
            remove_columns=["input", "output", "instruction", "data_source"],
            num_proc=self.data_args.preprocessing_num_workers,
            load_from_cache_file=not self.data_args.overwrite_cache,
            desc="Restructuring Platypus Dataset",
        )
        return raw_dataset



import time

from compressed_tensors.quantization.lifecycle.forward import (
    maybe_calibrate_or_quantize,
)

from llmcompressor.modifiers.utils.compression_wrapper import ModuleCompressionWrapper
from llmcompressor.utils import getattr_chain

try:
    import transformers
except ImportError as err:
    transformers = None
    transformers_err = err

import math

import torch
import torch.nn as nn
from loguru import logger

__all__ = ["SparseGptWrapper"]


class SparseGptWrapper(ModuleCompressionWrapper):
    """
    Runs SparseGPT on a single module that contains no sub-modules

    Lifecycle:
        - add_batch
        - compress
        - free

    :param name: name of module to run compression on
    :param layer: module to run compression on
    """

    def __init__(self, name, layer):
        super().__init__(name=name, layer=layer)

        # for Hessian calculation
        self.register_buffer(
            "H", torch.zeros((self.columns, self.columns), device=self.dev)
        )

    def add_batch(self, inp: torch.Tensor, out: torch.Tensor):
        """
        Add a batch of layer input and output data to the Hessian calculation

        :param inp: tensor containing layer input
        :param out: tensor containing layer output
        """
        if len(inp.shape) == 2:
            inp = inp.unsqueeze(0)
        tmp = inp.shape[0]
        if isinstance(self.layer, nn.Linear) or isinstance(
            self.layer, transformers.Conv1D
        ):
            if len(inp.shape) == 3:
                inp = inp.reshape((-1, inp.shape[-1]))
            inp = inp.t()
        self.H *= self.nsamples / (self.nsamples + tmp)
        self.nsamples += tmp
        inp = math.sqrt(2 / self.nsamples) * inp.float()
        self.H += inp.matmul(inp.t()).to(self.dev)

    def compress(
        self,
        sparsity: float,
        prunen: int = 0,
        prunem: int = 0,
        blocksize: int = 128,
        percdamp: float = 0.01,
        preserve_sparsity_mask: bool = False,
    ):
        """
        Run pruning and quantization(if applicable) on the layer up to the target
        sparsity value.

        :param sparsity: target sparsity to reach for layer
        :param prunen: N for N:M pruning
        :param prunem: M for N:M pruning
        :param blocksize: Number of columns to compress in one pass
        :param percdamp: Amount of dampening to apply to H, as a fraction of the
            diagonal norm
        :param preserve_sparsity_mask: Extend or ignore the base sparsity mask
        """
        final_shape = self.layer.weight.shape
        final_dtype = self.layer.weight.dtype
        W = self.layer.weight.data.clone()
        args_loc = "quantization_scheme.weights"
        weight_quant_args = getattr_chain(self.layer, args_loc, None)
        if weight_quant_args is not None:
            W = maybe_calibrate_or_quantize(self.layer, W, "weight", weight_quant_args)

        if isinstance(self.layer, nn.Conv2d):
            W = W.flatten(1)
        if isinstance(self.layer, transformers.Conv1D):
            W = W.t()
        W = W.float()

        tick = time.time()

        dead = torch.diag(self.H) == 0
        self.H[dead, dead] = 1
        W[:, dead] = 0

        Losses = torch.zeros(self.rows, device=self.dev)

        damp = percdamp * torch.mean(torch.diag(self.H))
        diag = torch.arange(self.columns, device=self.dev)
        self.H[diag, diag] += damp
        self.H = torch.linalg.cholesky(self.H)
        self.H = torch.cholesky_inverse(self.H)
        self.H = torch.linalg.cholesky(self.H, upper=True)
        Hinv = self.H

        mask = None
        if preserve_sparsity_mask:
            # compute existing sparsity mask
            mask = torch.where(
                W == 0,
                torch.tensor(1, dtype=torch.bool),
                torch.tensor(0, dtype=torch.bool),
            )
            current_sparsity = mask.sum() / W.numel()
            if current_sparsity > sparsity:
                raise ValueError(
                    "The target sparsity is lower than the sparsity "
                    "of the base model. Please retry "
                    "after turning preserve_sparsity_mask=False"
                )

        # See section 3.4 of https://arxiv.org/abs/2203.07259
        for i1 in range(0, self.columns, blocksize):
            i2 = min(i1 + blocksize, self.columns)
            count = i2 - i1

            W1 = W[:, i1:i2].clone()
            Q1 = torch.zeros_like(W1)
            Err1 = torch.zeros_like(W1)
            Losses1 = torch.zeros_like(W1)
            Hinv1 = Hinv[i1:i2, i1:i2]

            if prunen == 0:
                if mask is not None:
                    mask1 = mask[:, i1:i2]
                    if int(W1.numel() * sparsity) > mask1.sum():
                        # target sparsity is higher than base sparsity, extend mask1
                        tmp = (
                            (~mask[:, i1:i2])
                            * W1**2
                            / (torch.diag(Hinv1).reshape((1, -1))) ** 2
                        )
                        thresh = torch.sort(tmp.flatten())[0][
                            int(tmp.numel() * sparsity)
                        ]
                        mask1 = tmp <= thresh
                else:
                    tmp = W1**2 / (torch.diag(Hinv1).reshape((1, -1))) ** 2
                    thresh = torch.sort(tmp.flatten())[0][int(tmp.numel() * sparsity)]
                    mask1 = tmp <= thresh
            else:
                if mask is not None:
                    mask1 = mask[:, i1:i2]
                else:
                    mask1 = torch.zeros_like(W1) == 1

            for i in range(count):
                w = W1[:, i]
                d = Hinv1[i, i]

                if prunen != 0 and i % prunem == 0:
                    tmp = (
                        W1[:, i : (i + prunem)] ** 2
                        / (torch.diag(Hinv1)[i : (i + prunem)].reshape((1, -1))) ** 2
                    )
                    if mask is not None:
                        tmp = tmp * (~mask[:, i : (i + prunem)])

                    mask1.scatter_(
                        1, i + torch.topk(tmp, prunen, dim=1, largest=False)[1], True
                    )

                q = w.clone()
                q[mask1[:, i]] = 0

                Q1[:, i] = q
                Losses1[:, i] = (w - q) ** 2 / d**2

                err1 = (w - q) / d
                W1[:, i:] -= err1.unsqueeze(1).matmul(Hinv1[i, i:].unsqueeze(0))
                Err1[:, i] = err1

            W[:, i1:i2] = Q1
            Losses += torch.sum(Losses1, 1) / 2

            if preserve_sparsity_mask:
                # respect the sparsity of other groups
                # really not needed, but kept for explicitness
                W[:, i2:] -= (~mask[:, i2:]) * Err1.matmul(Hinv[i1:i2, i2:])
            else:
                W[:, i2:] -= Err1.matmul(Hinv[i1:i2, i2:])

        logger.info("time %.2f" % (time.time() - tick))
        logger.info("error %.2f" % torch.sum(Losses).item())

        if isinstance(self.layer, transformers.Conv1D):
            W = W.t()
        W = W.reshape(final_shape).to(final_dtype)
        if weight_quant_args is not None:
            W = maybe_calibrate_or_quantize(self.layer, W, "weight", weight_quant_args)

        # This is a bit hacky, but FSDP updates only work if we change the weight in
        # place, clone() or direct assignment won't work
        self.layer.weight -= self.layer.weight
        self.layer.weight += W

    def free(self):
        """
        Free the Hessian memory after the layer is complete
        """
        delattr(self, "H")
        super().free()



import math
import os
import re
from typing import List, Optional

import torch
from loguru import logger
from torch.utils.data import Dataset
from transformers import AutoTokenizer

from llmcompressor.core import active_session
from llmcompressor.pytorch.model_load.helpers import (
    get_completed_stages,
    get_session_model,
    save_completed_stages,
    save_model_and_recipe,
)
from llmcompressor.pytorch.utils import tensors_to_device
from llmcompressor.recipe import Recipe, StageRunType
from llmcompressor.transformers.finetune.data import TextGenerationDataset
from llmcompressor.transformers.finetune.data.data_args import DataTrainingArguments
from llmcompressor.transformers.finetune.data.data_helpers import (
    format_calibration_data,
    make_dataset_splits,
)
from llmcompressor.transformers.finetune.model_args import ModelArguments
from llmcompressor.transformers.finetune.training_args import TrainingArguments
from llmcompressor.utils.fsdp.helpers import (
    find_and_move_state_dicts_to_cpu,
    is_fsdp_model,
    unwrap_and_export_model,
)


class StageRunner:
    """
    Launcher class for train, eval and one_shot flows. Manages data splits for each
    flow and configurations. In the future this class will also handle alternating
    between the different flows

    LifeCycle
        - populate_datasets()
        - set_trainer()
        - train() / evaluate() / predict()

    :param model_args: Arguments pertaining to model/config/tokenizer
    :param data_args: Arguments pertaining to what data to use for different flows
    :param training_args: Arguments pertaining to training loop configuration
    :model: unwrapped model to run flows on
    """

    def __init__(
        self,
        data_args: "DataTrainingArguments",
        model_args: "ModelArguments",
        training_args: "TrainingArguments",
    ):
        self._data_args = data_args
        self._model_args = model_args
        self._training_args = training_args

        self.datasets = {}
        self.trainer = None
        self.tokenizer = None
        self.parent_output_dir = self._training_args.output_dir
        self._output_dir = self._training_args.output_dir

    def populate_datasets(self, tokenizer: "AutoTokenizer", add_labels: bool = True):
        """
        Loads datasets for each flow based on data_args, stores a Dataset for each
        enabled flow in self.datasets

        :param tokenizer: tokenizer to use for dataset tokenization
        """
        if self._data_args.dataset is None:
            self.tokenizer = self._model_args.tokenizer
            logger.info(
                "Running oneshot without calibration data. This is expected for "
                "weight-only and dynamic quantization"
            )
            return

        splits = self._data_args.splits
        tokenized_datasets = {}

        def _get_split_name(inp_str):
            # strip out split name, for ex train[60%:] -> train
            match = re.match(r"(\w*)\[.*\]", inp_str)
            if match is not None:
                return match.group(1)
            return inp_str

        if splits is None:
            splits = {"all": None}
        elif isinstance(splits, str):
            splits = {_get_split_name(splits): splits}
        elif isinstance(splits, List):
            splits = {_get_split_name(s): s for s in splits}

        # default to custom dataset if dataset provided isn't a string
        registry_id = self._data_args.dataset

        if not isinstance(registry_id, str):
            registry_id = "custom"
        for split_name, split_str in splits.items():
            dataset_manager = TextGenerationDataset.load_from_registry(
                registry_id,
                data_args=self._data_args,
                split=split_str,
                tokenizer=tokenizer,
            )

            dataset = self._data_args.dataset
            if hasattr(dataset, "column_names") and "input_ids" in dataset.column_names:
                # dataset is already tokenized
                tokenized_datasets[split_name] = dataset
            else:
                # dataset needs to be tokenized
                raw_dataset = dataset_manager.get_raw_dataset()
                tokenized_dataset = dataset_manager.tokenize_and_process(
                    raw_dataset, add_labels=add_labels
                )
                tokenized_datasets[split_name] = tokenized_dataset

        self.datasets = make_dataset_splits(
            tokenized_datasets,
            do_train=self._training_args.do_train,
            do_eval=self._training_args.do_eval,
            do_predict=self._training_args.do_predict,
            do_oneshot=self._training_args.do_oneshot,
        )
        self.tokenizer = tokenizer

    def get_dataset_split(self, split_name: str) -> Dataset:
        """
        Retrieve a dataset split by name

        :param split_name: name of dataset split to return
        :return: dataset split labeled by split_name
        """
        return self.datasets.get(split_name, None)

    def one_shot(self, stage: Optional[str] = None):
        """
        Run oneshot calibration on the active model

        :param stage: which stage of the recipe to run, or None to run whole recipe
        """
        logger.info("*** One Shot ***")

        calib_data = None
        if self.get_dataset_split("calibration") is not None:
            calib_data = format_calibration_data(
                tokenized_dataset=self.get_dataset_split("calibration"),
                num_calibration_samples=self._data_args.num_calibration_samples,
                do_shuffle=self._data_args.shuffle_calibration_samples,
                accelerator=self.trainer.accelerator,
            )

            # if we don't run a forward pass after initializing the FSDP model for the
            # first time, calls to summon_full_params will fail ¯\_(ツ)_/¯
            if is_fsdp_model(self.trainer.model):
                dummy_inp = dict(next(iter(calib_data)))
                model_device = next(self.trainer.model.parameters()).device
                dummy_inp = tensors_to_device(dummy_inp, model_device)
                with torch.no_grad():
                    self.trainer.model(**dummy_inp)

        self.trainer.accelerator.wait_for_everyone()

        self.trainer.one_shot(calibration_data=calib_data, stage=stage)

        if is_fsdp_model(self.trainer.model):
            try:
                self.trainer.save_model(output_dir=self._output_dir, _is_oneshot=True)
            except AssertionError:
                # fallback to this in the case of quantization
                unwrap_and_export_model(
                    model=self.trainer.model,
                    accelerator=self.trainer.accelerator,
                    output_dir=self._output_dir,
                    tokenizer=self.tokenizer,
                )
                # only allow the main process move the state
                # dicts to cpu
                if self.trainer.accelerator.is_main_process:
                    # assuming quantization is the last step
                    # we no longer need the original model
                    # and can safely delete it to save memory
                    del self.trainer.model
                    find_and_move_state_dicts_to_cpu(self._output_dir)

        else:
            save_model_and_recipe(
                model=self.trainer.model,
                save_path=self._output_dir,
                tokenizer=self.tokenizer,
                save_safetensors=self._training_args.save_safetensors,
                save_compressed=self._training_args.save_compressed,
            )

    def train(self, checkpoint: str, stage: Optional[str] = None):
        """
        Run trainer's training loop on train_dataset, saving the resulting model to
        output_dir

        :param checkpoint: Optional checkpoint to resume from
        :param stage: which stage of the recipe to run, or None to run whole recipe
        """
        logger.info("*** Train ***")
        train_result = self.trainer.train(
            resume_from_checkpoint=checkpoint, stage=stage
        )
        metrics = train_result.metrics
        metrics["train_samples"] = len(self.get_dataset_split("train"))
        metrics["perplexity"] = math.exp(metrics["train_loss"])
        self.trainer.log_metrics("train", metrics)
        self.trainer.save_metrics("train", metrics)

        # this includes saving the state, optimizer and scheduler
        self.trainer.save_model(output_dir=self._output_dir)

    def evaluate(self):
        """
        Run trainer's evaluation loop on eval_dataset, logging the desired metrics
        """
        logger.info("*** Evaluate ***")
        metrics = self.trainer.evaluate(self.get_dataset_split("validation"))

        metrics["eval_samples"] = len(self.get_dataset_split("validation"))
        self.trainer.log_metrics("eval", metrics)
        self.trainer.save_metrics("eval", metrics)

    def predict(self):
        """
        Run trainer's prediction loop on predict_dataset, logging the desired metrics
        """
        logger.info("*** Predict ***")
        results = self.trainer.predict(self.dataset["test"])
        metrics = results.metrics

        metrics["predict_samples"] = len(self.dataset["test"])
        self.trainer.log_metrics("predict", metrics)
        self.trainer.save_metrics("predict", metrics)

    def run_sequential_stages(self, checkpoint: Optional[str] = None):
        """
        Run the recipe stage by stage, allowing for alternating between one-shot and
        finetuning flows. Optionally save the model output at the end of each stage

        :param checkpoint: optional checkpoint to pick up a stage from
        """

        recipe_obj = Recipe.create_instance(self._training_args.recipe)
        with self.trainer.accelerator.main_process_first():
            checkpoint_dir = self._model_args.model
            completed_stages = get_completed_stages(checkpoint_dir)

        self.trainer.accelerator.wait_for_everyone()

        for stage in recipe_obj.stages:
            # validate stage
            stage_name = stage.group
            run_type = stage.infer_run_type()
            if not run_type:
                raise ValueError(
                    f"a valid stage type ({[e.value for e in StageRunType]}) "
                    "must be provided in run_stages mode. Either add a run_type "
                    "attribute to each stage in the recipe or include it as part of "
                    "the stage name."
                )

            # just load structure if stage has already applied
            if stage_name in completed_stages:
                self.trainer.initialize_structure(stage=stage)
                self.trainer.accelerator.wait_for_everyone()
                continue

            # setup checkpoint dir, TODO: this should be optional
            self._output_dir = os.path.join(
                self.parent_output_dir, "stage_" + stage_name
            )
            with self.trainer.accelerator.main_process_first():
                if not os.path.exists(self._output_dir):
                    os.makedirs(self._output_dir)
                save_completed_stages(self._output_dir, completed_stages)
            self._training_args.output_dir = self._output_dir

            # run stage
            if run_type is StageRunType.ONESHOT:
                self.one_shot(stage=stage_name)
            elif run_type is StageRunType.TRAIN:
                self.train(checkpoint=checkpoint, stage=stage_name)
            checkpoint = None

            # save stage stage to checkpoint dir
            if self.trainer.accelerator.is_main_process:
                completed_stages.append(stage_name)
                save_completed_stages(self._output_dir, completed_stages)

            # setup for next stage
            session = active_session()
            session.reset_stage()

            # synchronize and clean up memory
            self.trainer.accelerator.wait_for_everyone()
            self.trainer.model = get_session_model()
            torch.cuda.empty_cache()
            self.trainer.accelerator.free_memory()
            self.trainer.accelerator.wait_for_everyone()



from abc import ABC
from typing import Optional, Set

import torch
import torch.nn as nn
from torch.nn import Module

try:
    import transformers
except ImportError as err:
    transformers = None
    transformers_err = err

torch.backends.cuda.matmul.allow_tf32 = False
torch.backends.cudnn.allow_tf32 = False

__all__ = ["ModuleCompressionWrapper"]


class ModuleCompressionWrapper(Module, ABC):
    """
    Base Abstract class for pruning/quantization a single module with no sub-modules
    using information from input/output statistics

    Lifecycle:
        - add_batch
        - compress
        - free

    :param name: name of module to run compression on
    :param layer: module to run compression on
    """

    def __init__(self, name, layer):
        super(ModuleCompressionWrapper, self).__init__()
        if transformers is None:
            raise transformers_err

        self.name = name
        self.layer = layer

        self.dev = self.layer.weight.device
        if hasattr(self.layer, "_hf_hook") and self.layer._hf_hook.offload:
            self.dev = self.layer._hf_hook.execution_device

        # Calculate weight shape to use during pruning
        W = self.layer.weight
        if isinstance(self.layer, nn.Conv2d):
            self.rows = W.shape[0]
            self.columns = W.numel() / self.rows
        if isinstance(self.layer, transformers.Conv1D):
            self.rows = W.shape[1]
            self.columns = W.shape[0]
        else:
            self.rows = W.shape[0]
            self.columns = W.shape[1]

        # This need to be a buffer so its preserved between forward passes
        self.register_buffer(
            "nsamples", torch.zeros(1, dtype=torch.int32, device=self.dev)
        )

    def forward(self, *args, **kwargs):
        """
        Run a forward pass of the wrapped layer
        """
        return self.layer(*args, **kwargs)

    def free(self):
        """
        Free buffers used for compression
        """
        delattr(self, "nsamples")

    def add_batch(self, *args, **kwargs):
        """
        Add a batch of layer input and output data to the layer statistics calculation
        """
        raise NotImplementedError("Child class must implement `add_batch`")

    def compress(self, *args, **kwargs):
        """
        Run pruning on the layer up to the target sparsity
        """
        raise NotImplementedError("Child class must implement `compress`")

    def state_dict(self, destination=None, prefix="", keep_vars=False, **kwargs):
        """
        Pass request to wrapped layer, so compression wrapper does not appear in
        the state_dict
        """
        return self.layer.state_dict(
            destination=destination, prefix=prefix, keep_vars=keep_vars, **kwargs
        )

    def load_state_dict(self, state_dict, strict=True):
        """
        Pass request to wrapped layer, so compression wrapper does not appear in
        the state_dict
        """
        return self.layer.load_state_dict(state_dict, strict=strict)

    def named_modules(
        self,
        memo: Optional[Set["Module"]] = None,
        prefix: str = "",
        remove_duplicate: bool = True,
    ):
        """
        Pass request to wrapped layer, so compression wrapper does not appear in
        the module list
        """
        return self.layer.named_modules(
            memo=memo, prefix=prefix, remove_duplicate=remove_duplicate
        )



from copy import deepcopy

from llmcompressor.transformers.finetune.data import TextGenerationDataset


@TextGenerationDataset.register(name="ptb")
class PtbDataset(TextGenerationDataset):
    """
    Child text generation class for the PTB dataset

    :param data_args: configuration settings for dataset loading
    :param split: split from dataset to load, for instance `test` or `train[:5%]`
    :param tokenizer: tokenizer to use on dataset
    """

    def __init__(self, data_args, split, tokenizer):
        data_args = deepcopy(data_args)
        data_args.dataset = "ptb_text_only"
        super().__init__(
            text_column="sentence",
            data_args=data_args,
            split=split,
            tokenizer=tokenizer,
        )
